from unittest.mock import Mock, patch

import pytest

import toddlers.utils.fedora_account


class TestFedoraAccount:
    def test_set_fas_no_fas_url(self):
        with pytest.raises(
            ValueError, match=r"No fas_url found in the configuration file"
        ):
            toddlers.utils.fedora_account.set_fas({})

    def test_set_fas_no_fas_username(self):
        with pytest.raises(
            ValueError, match=r"No fas_username found in the configuration file"
        ):
            config = {
                "fas_url": "https:fas.example.com",
            }
            toddlers.utils.fedora_account.set_fas(config)

    def test_set_fas_no_fas_password(self):
        with pytest.raises(
            ValueError, match=r"No fas_password found in the configuration file"
        ):
            config = {
                "fas_url": "https:fas.example.com",
                "fas_username": "fas_user",
            }
            toddlers.utils.fedora_account.set_fas(config)

    @patch("toddlers.utils.fedora_account.AccountSystem")
    def test_set_fas(self, mock_fas):
        mock_fas.return_value = "fas_object"
        config = {
            "fas_url": "https:fas.example.com",
            "fas_username": "fas_user",
            "fas_password": "fas_password",
        }
        output = toddlers.utils.fedora_account.set_fas(config)
        mock_fas.assert_called_with(
            "https:fas.example.com",
            username="fas_user",
            password="fas_password",
            cache_session=False,
        )
        assert output == "fas_object"

    @patch("toddlers.utils.fedora_account._FAS", new=None)
    def test_get_fas_not_set(self):
        with pytest.raises(
            ValueError, match=r"No FAS connection set, call set_fas first"
        ):
            toddlers.utils.fedora_account.get_fas()

    def test_get_fas(self):
        output = toddlers.utils.fedora_account.get_fas()
        assert output == "fas_object"

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_group_member(self, mock_fas):
        members = []
        for name in ["pingou", "ralph", "kevin", "nils"]:
            member = Mock()
            member.role_type = "administrator"
            member.username = name
            members.append(member)
        server = Mock()
        server.group_members.return_value = members
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_group_member("sysadmin")
        assert output == {"kevin", "nils", "pingou", "ralph"}

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_group_member_empty(self, mock_fas):
        members = []
        server = Mock()
        server.group_members.return_value = members
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_group_member("sysadmin")
        assert output == set()

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_bz_email_user_no_bugzilla_email(self, mock_fas):
        server = Mock()
        server.person_by_username.return_value = {}
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_bz_email_user("pingou", {})
        assert output is None

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_bz_email_user(self, mock_fas):
        server = Mock()
        server.person_by_username.return_value = {"bugzilla_email": "foo@bar.com"}
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_bz_email_user("pingou", {})
        assert output == "foo@bar.com"

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_bz_email_user_overriden(self, mock_fas):
        server = Mock()
        server.person_by_username.return_value = {"bugzilla_email": "foo@bar.com"}
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_bz_email_user(
            "pingou", {"foo@bar.com": "foo@baz.com"}
        )
        assert output == "foo@baz.com"

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_bz_email_group_no_bugzilla_email(self, mock_fas):
        server = Mock()
        server.group_by_name.return_value = {}
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_bz_email_group("toddlers-sig", {})
        assert output is None

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_bz_email_group(self, mock_fas):
        server = Mock()
        server.group_by_name.return_value = {"mailing_list": "foo@lists.bar.com"}
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_bz_email_group("toddlers-sig", {})
        assert output == "foo@lists.bar.com"

    @patch("toddlers.utils.fedora_account.get_fas")
    def test_get_bz_email_group_overriden(self, mock_fas):
        server = Mock()
        server.group_by_name.return_value = {"mailing_list": "foo@lists.bar.com"}
        mock_fas.return_value = server

        output = toddlers.utils.fedora_account.get_bz_email_group(
            "toddlers-sig", {"foo@lists.bar.com": "foo@baz.com"}
        )
        assert output == "foo@baz.com"
