"""
This plugin handles fedora-messaging messages published under the topic
``ci.pipeline.allpackages-pr.complete/running/error`` and flag the
corresponding PR in dist-git with the result of the tests.

Authors:    Pierre-Yves Chibon <pingou@pingoured.fr>

"""

import hashlib
import logging

import requests
from requests.packages.urllib3.util import retry

from ..base import ToddlerBase


_log = logging.getLogger(__name__)

timeout = (30, 30)
retries = 3
requests_session = requests.Session()
retry_conf = retry.Retry(total=retries, connect=retries, read=retries, backoff_factor=1)
retry_conf.BACKOFF_MAX = 5
requests_session.mount("http://", requests.adapters.HTTPAdapter(max_retries=retry_conf))
requests_session.mount(
    "https://", requests.adapters.HTTPAdapter(max_retries=retry_conf)
)

done_states = {
    "passed": {"api": "success", "human": "passed"},
    "error": {"api": "error", "human": "errored"},
    "failed": {"api": "failure", "human": "failed"},
    "needs_inspection": {"api": "failure", "human": "failed"},
}


class FlagCIPR(ToddlerBase):
    """ Listens to messages sent by fedora CI and flags pull-requests on
    src.fedoraproject.org accordingly.
    """

    name = "flag_ci_pr"

    amqp_topics = [
        "org.centos.*.ci.dist-git-pr.test.error",
        "org.centos.*.ci.dist-git-pr.test.complete",
        "org.centos.*.ci.dist-git-pr.test.running",
    ]

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return (
            topic.startswith("org.centos.")
            and "ci.dist-git-pr.test" in topic
            and topic.endswith(("complete", "error", "running"))
        )

    def process(self, config, message):
        """Process a given message."""

        pipeline_state = None
        if "." in message.topic:
            pipeline_state = message.topic.rsplit(".", 1)[-1]

        if pipeline_state not in ["complete", "running", "error"]:
            _log.info("Pipeline state is not 'complete' or 'running' or 'error'.")
            return

        msg = message.body

        if "version" not in msg or not msg["version"].startswith("0.2."):
            _log.info("Unsupported msg version, ignoring")
            return

        # test complete messages
        if pipeline_state == "complete":
            state = msg["test"]["result"]
            if state not in done_states:
                _log.info("Build is not in one of the expected status, ignoring")
                return

            status = done_states[state]["api"]
            human_status = done_states[state]["human"]

        # test running messages
        elif pipeline_state == "running":
            status = "pending"
            human_status = "running"

        # test error messages
        elif pipeline_state == "error":
            status = "error"
            human_status = "errored"

        if "id" not in msg["artifact"]:
            _log.info(
                "Invalid message: %s, could not extract the PR id from it"
                % msg["artifact"]
            )
            return

        pr_id = str(msg["artifact"]["id"])

        commit_hash_text = " for %s" % msg["artifact"]["commit_hash"][:8]
        seed = config.get("pagure_token_seed") or ""

        data = {
            "username": "Fedora CI",
            "status": status,
            "comment": "Package tests%s: %s" % (commit_hash_text, human_status),
            "url": msg["run"]["url"],
            "uid": hashlib.md5((pr_id + seed).encode("utf-8")).hexdigest(),
        }

        pagure_url = config["pagure_url"]

        namespace = msg["artifact"]["repository"].split("/")[-2]
        repo = msg["artifact"]["repository"].split("/")[-1]

        target_url = "/".join(
            ["api", "0", namespace, repo, "pull-request", pr_id, "flag"]
        )
        flag_url = pagure_url + "/" + target_url
        _log.info("Flagging commit at: %s" % flag_url)

        headers = {
            "Authorization": "token " + config["pagure_token"],
            "User-Agent": "loopabull@fedora-infra",
        }

        _log.info("payload: %s" % data)

        req = requests_session.request(
            method="POST", url=flag_url, headers=headers, data=data,
        )
        _log.info("Request to %s returned: %s" % (pagure_url, req.status_code))
        if not req.ok:
            _log.debug(req.text)
            return 1
        else:
            _log.info("All clear")
            _log.info(
                "User-URL: %s/" % msg["artifact"]["repository"]
                + "/".join(["pull-request", pr_id])
            )
