"""
When a component's branch is retired, EOL all SLAs on that branch.

This toddler is succeeding to the pdc-updater handlers:
https://github.com/fedora-infra/pdc-updater/blob/develop/pdcupdater/handlers/retirement.py

Authors:    Pierre-Yves Chibon <pingou@pingoured.fr>

"""

from datetime import datetime
import logging

import pdc_client
import requests
from requests.packages.urllib3.util import retry

from toddlers.base import ToddlerBase

_log = logging.getLogger(__name__)

timeout = (30, 30)
retries = 3
requests_session = requests.Session()
retry_conf = retry.Retry(total=retries, connect=retries, read=retries, backoff_factor=1)
retry_conf.BACKOFF_MAX = 5
requests_session.mount("http://", requests.adapters.HTTPAdapter(max_retries=retry_conf))
requests_session.mount(
    "https://", requests.adapters.HTTPAdapter(max_retries=retry_conf)
)


class PDCRetiredPackages(ToddlerBase):
    """ Listens to messages sent by playtime (which lives in toddlers) to sync
    all the packagers found in FAS to bugzilla.
    """

    name = "pdc_retired_packages"

    amqp_topics = [
        "org.fedoraproject.*.toddlers.trigger.pdc_retired_packages",
        "org.fedoraproject.*.git.receive",
    ]

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            ("toddlers.trigger.pdc_retired_packages", "git.receive")
        )

    def process(self, config, message):
        """Process a given message."""
        pdc = pdc_client.PDCClient(**config.get("pdc_config", {}))

        if message.topic.endswith("git.receive"):
            PDCRetiredPackages._process_single_package(config, pdc, message)
        else:
            PDCRetiredPackages._process_dist_git(config, pdc)

    def _process_dist_git(self, config, pdc):
        """ Updates PDC retirement status from analyzing dist-git.

        This steps over all the branches in dist-git and retires any branches
        in PDC that have a dead.package file in dist-git.
        """
        _log.info("Looking up all branches from PDC.")
        for branch in pdc.get_paged(pdc["component-branches"]._):
            branch_str = "{type}/{global_component}#{name}".format(**branch)
            _log.debug("Considering {0}".format(branch_str))
            retired_in_dist_git = False
            try:
                retired_in_dist_git = _is_retired_in_dist_git(
                    namespace=_pdc_to_namespace(branch["type"]),
                    repo=branch["global_component"],
                    branch=branch["name"],
                )
            except ValueError as err:
                _log.debug("Ignoring PDC error for: %s", err)
                continue

            if retired_in_dist_git:
                _retire_branch(pdc, branch)

    def _process_single_package(self, config, pdc, message):
        """Handle an incoming bus message.

        The message should be a dist-git retirement message (where someone adds
        a dead.package file to the repo).  In response, this method will retire
        the package in PDC.
        """

        msg = message.body

        # If there is no dead.package in the commit, then it can be ignored
        if "dead.package" not in msg["commit"]["stats"]["files"]:
            _log.info("No dead.package in the commit, bailing")
            return

        dead_package_commit = msg["commit"]["stats"]["files"]["dead.package"]
        if (
            dead_package_commit["additions"] == 0
            and dead_package_commit["deletions"] != 0
        ):
            _log.info("dead.package file was not added, bailing")
            return

        branchname = msg["commit"]["branch"]
        repo = msg["commit"]["repo"]
        namespace = msg["commit"]["namespace"]
        try:
            component_type = _namespace_to_pdc(namespace)
        except ValueError as err:
            _log.info("Ignoring namespace '%s', error: %s", namespace, err)
            return

        checkurl = config.get("file_check_url")
        if not checkurl:
            _log.error("No check URL configured, ignoring")
            return

        # This query guarantees a unique component branch, so a count of 1 is
        # expected
        branch_query_rv = pdc["component-branches"]._(
            name=branchname, type=component_type, global_component=repo
        )

        if branch_query_rv["count"] != 1:
            _log.error('"{0}/{1}" was not found in PDC'.format(namespace, repo))
            return

        branch = branch_query_rv["results"][0]
        # If the branch is already EOL in PDC, don't do anything
        if branch["active"] is False:
            _log.error(
                '"{0}/{1}" not active in PDC in {2}'.format(
                    namespace, repo, branch["name"]
                )
            )
            return

        # Make sure that the file is still retired right this moment.
        # This fixes cases where people merge from master to an older branch, and one of the
        # intermediate commits contained a dead.package.
        fileurl = checkurl % {
            "namespace": namespace,
            "repo": repo,
            "branch": branchname,
            "file": "dead.package",
        }
        _log.info("Checking for file: %s", fileurl)
        resp = requests.head(fileurl, timeout=15)
        if resp.status_code != 200:
            _log.info("Seems not to actually be retired, possibly a merge commit?")
            return

        _retire_branch(pdc, branch)


def _namespace_to_pdc(namespace):
    """ Internal method to translate a dist-git namespace to a PDC
    component type. """
    namespace_to_pdc = {
        "rpms": "rpm",
        "modules": "module",
        "container": "container",
    }
    if namespace not in namespace_to_pdc:
        raise ValueError('The namespace "{0}" is not supported'.format(namespace))
    else:
        return namespace_to_pdc[namespace]


def _pdc_to_namespace(pdc_type):
    """ Internal method to translate a PDC component type to a dist-git
    namespace. """
    pdc_to_namespace = {
        "rpm": "rpms",
        "module": "modules",
        "container": "container",
    }
    if pdc_type not in pdc_to_namespace:
        raise ValueError('The PDC type "{0}" is not supported'.format(pdc_type))
    else:
        return pdc_to_namespace[pdc_type]


def _is_retired_in_dist_git(namespace, repo, branch):
    base = "https://src.fedoraproject.org/"
    # Check to see if they have a dead.package file in dist-git
    url = "{base}/{namespace}/{repo}/raw/{branch}/f/dead.package"
    response = requests_session.head(
        url.format(base=base, namespace=namespace, repo=repo, branch=branch,)
    )

    # If there is a dead.package, then the branch is retired in dist_git
    if response.status_code in [200, 404]:
        return response.status_code == 200
    else:
        raise ValueError(
            "The connection to dist_git failed. Retirement status could not "
            "be determined. The status code was: {0}. The content was: "
            "{1}".format(response.status_code, response.content)
        )


def _retire_branch(pdc, branch):
    """ Internal method for retiring a branch in PDC. """
    today = datetime.utcnow().date()
    for sla in branch["slas"]:
        sla_eol = datetime.strptime(sla["eol"], "%Y-%m-%d").date()
        if sla_eol > today:
            _log.info("Setting eol of branch: %s to %s", branch, today)
            pdc["component-branch-slas"][sla["id"]]._ += {"eol": str(today)}
