import logging

import fedora_messaging.config
import fedora_messaging.exceptions

from . import plugins  # noqa: F401
from .base import ToddlerBase


_log = logging.getLogger(__name__)


class RunningToddler(object):
    """
    The fedora-messaging consumer of the project.

    This consumer loads all the plugins implemeting the toddler.base.ToddlerBase
    interface and for every message it receives it asks them if they care to
    handle the message and if so let them process it.

    """

    def __init__(self):
        """Perform some one-time initialization for the consumer."""
        _log.debug("Loading toddlers")
        blocked_toddlers = (
            fedora_messaging.config.conf["consumer_config"].get("blocked_toddlers")
            or []
        )
        self.toddlers = []
        for toddler in ToddlerBase.__subclasses__():
            if toddler.name in blocked_toddlers:
                _log.info("Toddler '%s' is blocked, skipping it", toddler.name)
                continue
            self.toddlers.append(toddler())
        _log.info("Loaded: %s", self.toddlers)
        if not self.toddlers:
            raise fedora_messaging.exceptions.ConfigurationException(
                "No toddlers found to run :("
            )

        topics = set()
        for toddler in self.toddlers:
            _log.debug(
                "Toddler '%s' cares about the topics: %s",
                toddler.name,
                toddler.amqp_topics,
            )
            topics = topics.union(set(toddler.amqp_topics))
        _log.info("Topics of interest: %s", topics)

        for binding in fedora_messaging.config.conf["bindings"]:
            if binding["routing_keys"] == ["#"]:
                _log.info(
                    "Updating the routing_keys of the queue %s to the topics of interest",
                    binding["queue"],
                )
                binding["routing_keys"] = topics

    def __call__(self, message):
        """
        Invoked when a message is received by the consumer.

        Args:
            message (fedora_messaging.api.Message): The message from AMQP.
        """
        _log.info("Toddlers processing: %s -- %s", message.id, message.topic)
        for toddler in self.toddlers:
            _log.debug("Running toddler: %s", toddler.name)
            try:
                if toddler.accepts_topic(message.topic):
                    _log.info(
                        "Toddler '%s' accepted to process message id: %s",
                        toddler.name,
                        message.id,
                    )
                    toddler.process(
                        fedora_messaging.config.conf["consumer_config"].get(
                            toddler.name
                        ),
                        message,
                    )
            except Exception:
                _log.exception(
                    "Toddler '%s' failed to process message id: %s -- putting it back in the queue",
                    toddler.name,
                    message.id,
                )
                raise fedora_messaging.exceptions.Nack("A toddler tripped")
