import logging
from typing import Mapping
import xmlrpc.client

from bugzilla import Bugzilla


_log = logging.getLogger(__name__)
# Have a global connection to _BUGZILLA open.
_BUGZILLA = None


def set_bz(conf: Mapping[str, str]) -> Bugzilla:
    """ Set the connection bugzilla.
    """
    global _BUGZILLA

    # Get a connection to bugzilla
    bz_server = conf.get("bugzilla_url")
    if not bz_server:
        raise ValueError("No bugzilla_url found in the configuration file")

    bz_url = bz_server + "/xmlrpc.cgi"

    bz_user = conf.get("bugzilla_username")
    if not bz_user:
        raise ValueError("No bugzilla_username found in the configuration file")

    bz_pass = conf.get("bugzilla_password")
    if not bz_pass:
        raise ValueError("No bugzilla_password found in the configuration file")

    _BUGZILLA = Bugzilla(
        url=bz_url, user=bz_user, password=bz_pass, cookiefile=None, tokenfile=None
    )
    return _BUGZILLA


def get_bz() -> Bugzilla:
    """Retrieve a connection to bugzilla

    :raises xmlrpclib.ProtocolError: If we're unable to contact bugzilla
    """
    if _BUGZILLA is None:
        raise ValueError("No bugzilla connection set, call set_bz first")

    return _BUGZILLA


def get_group_member(group_name: str) -> list:
    """ Return a list containing the name the members of the given group.

    :arg group_name: The group name used in bugzilla.
    :raises XMLRPC Fault: Code 51 if the name does not exist
    :raises XMLRPC Fault: Code 805 if the user does not have enough
        permissions to view groups
    """

    server = get_bz()

    group = server.getgroup(group_name, membership=True)

    return group.member_emails


def remove_user_from_group(user_email: str, bz_group: str):
    """ Remove the specified user from the specified group. """

    server = get_bz()

    # Remove the user's bugzilla group
    try:
        server.updateperms(user_email, "rem", bz_group)
    except xmlrpc.client.Fault as e:
        if e.faultCode == 51:
            # It's okay, not having this user is equivalent to setting
            # them to not have this group.
            pass
        else:
            raise


def user_exists(user_email: str):
    """ Returns a boolean specifying if the given user exists in bugzilla or not. """

    server = get_bz()

    # Make sure the user exists
    try:
        server.getuser(user_email)
        output = True
        # print(user.userid, user.name, user.email, user.groupnames)
    except xmlrpc.client.Fault as e:
        if e.faultCode == 51:
            output = False
        else:
            raise
    return output


def add_user_to_group(
    user_email: str, bz_group: str, no_bz_account: list, dry_run: bool = False
):
    """ Add the specified user to the specified group. """

    server = get_bz()

    # Make sure the user exists
    if not user_exists(user_email):
        # This user doesn't have a bugzilla account yet
        # add them to a list and we'll let them know.
        no_bz_account.append(user_email)
        return no_bz_account

    if not dry_run:
        # Add the user to the group
        print(server.updateperms(user_email, "add", bz_group))
    else:
        _log.info("   Would add %s to the group %s", user_email, bz_group)

    return no_bz_account
