from typing import Any
from typing import Mapping

from fedora.client.fas2 import AccountSystem


# Have a global connection to FAS open.
_FAS = None


def set_fas(conf: Mapping[str, str]) -> AccountSystem:
    """ Set the connection to the Fedora Account System.
    """
    global _FAS

    # Get a connection to FAS
    fas_url = conf.get("fas_url")
    if not fas_url:
        raise ValueError("No fas_url found in the configuration file")

    fas_user = conf.get("fas_username")
    if not fas_user:
        raise ValueError("No fas_username found in the configuration file")

    fas_pass = conf.get("fas_password")
    if not fas_pass:
        raise ValueError("No fas_password found in the configuration file")

    _FAS = AccountSystem(
        fas_url, username=fas_user, password=fas_pass, cache_session=False
    )

    return _FAS


def get_fas() -> AccountSystem:
    """ Retrieve a connection to the Fedora Account System.
    """
    global _FAS
    if _FAS is None:
        raise ValueError("No FAS connection set, call set_fas first")

    return _FAS


def __get_fas_grp_member(group: str = "packager") -> Mapping[str, Mapping[str, Any]]:
    """ Retrieve from FAS the list of users in the packager group.
    """
    fas = get_fas()

    return fas.group_members(group)


def get_group_member(group_name: str) -> set:
    """ Return a list containing the name the members of the given group. """
    output = set()
    for user in __get_fas_grp_member(group_name):
        if user.role_type in ("user", "sponsor", "administrator"):
            output.add(user.username)
    return output


def get_bz_email_user(username, email_overrides):
    """ Retrieve the bugzilla email associated to the provided username.
    """
    fas = get_fas()

    user_info = fas.person_by_username(username)
    bz_email = user_info.get("bugzilla_email", None)
    if bz_email is None:
        return

    bz_email = bz_email.lower()
    bz_email = email_overrides.get(bz_email, bz_email)
    return bz_email


def get_bz_email_group(groupname, email_overrides):
    """ Retrieve the bugzilla email associated to the provided group name.
    """
    fas = get_fas()

    group = fas.group_by_name(groupname)
    bz_email = group.get("mailing_list")
    if bz_email is None:
        return

    bz_email = bz_email.lower()
    bz_email = email_overrides.get(bz_email, bz_email)
    return bz_email
